@@ -8,6 +8,7 @@ import dateutil.parser
 
 from imecilabt_utils.urn_util import URN, always_optional_urn
 from imecilabt_utils.validation_utils import is_valid_uuid4, is_valid_uuid
+from imecilabt.gpulab.util.convert_utils import urn_to_user_mini_id, urn_to_name, urn_to_auth
 
 NVDATA_JOBDEFINITION = "jobDefinition"
 
@@ -55,43 +56,6 @@ def _int_or_none(val: Optional[Any]) -> Optional[int]:
     return int(val) if val is not None else None
 
 
-def urn_to_user_mini_id(urn: Optional[Union[str, URN]]) -> Optional[str]:
-    if not urn:
-        return None
-    u = always_optional_urn(urn)
-    if u and u.authority and u.name:
-        auth = u.authority
-        if auth == 'wall2.ilabt.iminds.be':
-            auth = 'w2'
-        if auth == 'ilabt.imec.be':
-            auth = 'ilabt'
-        if auth == 'example.com':  # used in tests
-            auth = 'ex'
-        return '{name}@{auth}'.format(auth=auth, name=u.name)
-    else:
-        return None
-
-
-def urn_to_auth(urn: Optional[Union[str, URN]]) -> Optional[str]:
-    if not urn:
-        return None
-    u = always_optional_urn(urn)
-    if u and u.authority:
-        return u.authority
-    else:
-        return None
-
-
-def urn_to_name(urn: Optional[Union[str, URN]]) -> Optional[str]:
-    if not urn:
-        return None
-    u = always_optional_urn(urn)
-    if u and u.name:
-        return u.name
-    else:
-        return None
-
-
 class JobPortMapping:
     def __init__(self,
                  container_port: Optional[int] = None,
@@ -791,6 +755,7 @@ class Job:
         assert created is None or created.tzinfo is not None
         assert state_updated is None or state_updated.tzinfo is not None
 
+        assert uuid is None or isinstance(uuid, str)
         self.uuid = uuid
         self.id = uuid   #id and uuid are synonyms for Job
         self.name = name
