@@ -9,13 +9,15 @@ import dateutil.parser
 from stringcase import camelcase
 from imecilabt_utils.utils import duration_string_to_seconds
 
-from imecilabt.gpulab.model.job import urn_to_name, urn_to_auth, urn_to_user_mini_id, DOCKER_IMAGE_USERPASS_PATTERN
+from imecilabt.gpulab.model.job import DOCKER_IMAGE_USERPASS_PATTERN
 from imecilabt.gpulab.model.usage_statistics import GPULabUsageStatistics, GpuOverview
 from dataclass_dict_convert import dataclass_dict_convert, create_wrap_in_list_from_convertor, \
     dataclass_auto_type_check, dataclass_copy_method, dataclass_multiline_repr
 from imecilabt_utils.urn_util import URN, is_valid_urn, check_valid_urn_bytype
 from imecilabt_utils.validation_utils import is_valid_uuid, is_valid_email, is_valid_ssh_key, ALLOWED_HOST_KEY_ALGOS
 
+from imecilabt.gpulab.util.convert_utils import urn_to_user_mini_id, urn_to_name, urn_to_auth
+
 
 def is_valid_job_id(id: str) -> bool:
     return is_valid_uuid(id)  # we allow both version 1 and 4 UUIDs (version 4 is used for new jobs)
@@ -188,7 +190,29 @@ class JobEventTimes:
         return dataclasses.replace(self, long_run_notify=None)
 
 
-@dataclass_dict_convert(dict_letter_case=camelcase)
+def custom_tmpfs_mem_gb_to_dict(tmpfs_mem_gb: Optional[int]) -> Optional[int]:
+    if tmpfs_mem_gb is None or tmpfs_mem_gb == 0:
+        return None
+    else:
+        return tmpfs_mem_gb
+
+
+def custom_tmpfs_mem_gb_from_dict(tmpfs_mem_gb: Optional[int]) -> Optional[int]:
+    if tmpfs_mem_gb is None or tmpfs_mem_gb == 0:
+        return 0
+    else:
+        return tmpfs_mem_gb
+
+
+@dataclass_dict_convert(
+    dict_letter_case=camelcase,
+    custom_to_dict_convertors={
+        'tmpfs_mem_gb': custom_tmpfs_mem_gb_to_dict
+    },
+    custom_from_dict_convertors={
+        'tmpfs_mem_gb': custom_tmpfs_mem_gb_from_dict
+    }
+)
 @dataclass(frozen=True)
 @dataclass_auto_type_check
 @dataclass_copy_method
@@ -197,7 +221,7 @@ class JobStateResources:
     cluster_id: int
     gpu_ids: List[int]
     cpu_ids: List[int]
-    cpu_memory_gb: int
+    cpu_memory_gb: int  # does NOT include tmpfs memory (it is listed separately below)
     gpu_memory_gb: int  # if less than a GB, this is rounded DOWN (so it can be 0!)
 
     slave_name: str
@@ -216,6 +240,8 @@ class JobStateResources:
     port_mappings: List[JobPortMapping] = field(default_factory=list)
     gpu_details: Optional[GpuOverview] = None
 
+    tmpfs_mem_gb: Optional[int] = 0  # None = 0
+
     @property
     def cpu_memory_mb(self) -> int:
         # while GB is sometimes 1000^3 instead of 1024^3 (typically HD size etc), for memory, base 1024 is always used.
@@ -228,6 +254,34 @@ class JobStateResources:
         # (and docker seems to use 1000 based, which is silly)
         return self.cpu_memory_gb * 1024 * 1024 * 1024
 
+    @property
+    def tmpfs_mem_mb(self) -> int:
+        # while GB is sometimes 1000^3 instead of 1024^3 (typically HD size etc), for memory, base 1024 is always used.
+        # (and docker seems to use 1000 based, which is silly)
+        return (self.tmpfs_mem_gb or 0) * 1024
+
+    @property
+    def tmpfs_mem_byte(self) -> int:
+        # while GB is sometimes 1000^3 instead of 1024^3 (typically HD size etc), for memory, base 1024 is always used.
+        # (and docker seems to use 1000 based, which is silly)
+        return (self.tmpfs_mem_gb or 0) * 1024 * 1024 * 1024
+
+    @property
+    def all_cpu_memory_gb(self) -> int:
+        return self.cpu_memory_gb + (self.tmpfs_mem_gb or 0)
+
+    @property
+    def all_cpu_memory_mb(self) -> int:
+        # while GB is sometimes 1000^3 instead of 1024^3 (typically HD size etc), for memory, base 1024 is always used.
+        # (and docker seems to use 1000 based, which is silly)
+        return (self.all_cpu_memory_gb or 0) * 1024
+
+    @property
+    def all_cpu_memory_byte(self) -> int:
+        # while GB is sometimes 1000^3 instead of 1024^3 (typically HD size etc), for memory, base 1024 is always used.
+        # (and docker seems to use 1000 based, which is silly)
+        return (self.all_cpu_memory_gb or 0) * 1024 * 1024 * 1024
+
     def sanitized_copy(self, logged_in=True, same_project=False) -> 'JobStateResources':
         # not entirely public, but all not that private
         res = dataclasses.replace(
@@ -295,6 +349,19 @@ class JobRequestExtra:
 @dataclass_copy_method
 @dataclass_multiline_repr
 class JobRequestScheduling:
+    """
+    GPULab scheduling instructions.
+
+    Attributes:
+      interactive: Interactive jobs will either run immediately, or fail directly. They will never be QUEUED for a long time.
+      killable_after: GPULab might kill this job after this duration. Setting this allows GPULab to schedule this job earlier than it can otherwise, but there is a chance your job will be stopped after this time. (If it is restartable, it will however restart later.) Format: a number followed by a unit (ex: 5 minutes, 3 hour, 2 days, 1 week, ...)
+      restartable: Restartable jobs can be stopped and later restarted by GPULab. Before stopping, GPULab will send a signal to the job. In exchange for this flexibility, GPULab can start the jobs sooner than it otherwise would.
+      reservation_id: The reservation ID to use it for the job. This allows the job to start when it otherwise would not be able to start due to a reservation.
+      max_duration: The maximum duration of this job. GPULab will always stop your job after this time. If the job has been restarted once or more, the total duration of all the job runs is used. Format: a number followed by a unit (ex: 5 minutes, 3 hour, 2 days, 1 week, ...)
+      max_simultaneous_jobs: Control how many of your jobs can run at the same time. Format: { "bucket_name": <bucketname>, "bucket_max": <max number> }
+      not_before: Request GPULab not to start the job before a specified date. It will stay QUEUED at least until the requested time. Format: an RFC3339 date.
+      not_after: Request GPULab to FAIL the job if it is still QUEUED at a specified date. (Does not affect an already running job, only prevents start after the date.) Format: an RFC3339 date.
+    """
     interactive: bool = False
     killable_after: Optional[str] = None  # duration str
     restartable: bool = False
@@ -316,6 +383,12 @@ class JobRequestScheduling:
             if max_dur_s < 60:
                 raise AttributeError('max_duration must be at least 1 minute, not "{}"'
                                      .format(self.max_duration))
+        if self.not_before and self.interactive:
+            raise AttributeError('not_before cannot be combined with interactive')
+        if self.not_after and self.interactive:
+            raise AttributeError('not_after cannot be combined with interactive')
+        if self.not_before and self.not_after and self.not_before >= self.not_after:
+            raise AttributeError(f'not_before ({self.not_before}) must be before not_after ({self.not_after})')
 
     @property
     def max_duration_s(self) -> Optional[int]:
@@ -431,6 +504,7 @@ class JobRequestResources:
 class JobStorage:
     container_path: Optional[str] = None  # default: host_path
     host_path: Optional[str] = None  # default: container_path
+    size_gb: Optional[int] = None  # default: None
 
     def __post_init__(self):
         if not self.container_path and not self.host_path:
@@ -440,9 +514,17 @@ class JobStorage:
         if not self.host_path:
             object.__setattr__(self, 'host_path', self.container_path)
 
+    # only useful for backward compat testing
+    def _without_size(self) -> 'JobStorage':
+        return JobStorage(container_path=self.container_path, host_path=self.host_path)
+
+    @property
+    def is_tmpfs(self) -> bool:
+        return self.host_path == 'tmpfs'
+
     @classmethod
     def from_string(cls, s: str) -> 'JobStorage':
-        return cls(container_path=s, host_path=s)
+        return cls(container_path=s, host_path=s, size_gb=None)
 
 
 def _convert_portmappings(pm_list: List):
@@ -464,7 +546,7 @@ def _convert_storage(storage_list: List):
     res = []
     for s in storage_list:
         if isinstance(s, str):
-            res.append(JobStorage(container_path=s, host_path=s))
+            res.append(JobStorage(container_path=s, host_path=s, size_gb=None))
         else:
             assert isinstance(s, dict)
             res.append(JobStorage.from_dict(s))
@@ -528,6 +610,26 @@ class JobRequestDocker:
         _test_char_list(self.command, 'Job.request.docker.command', dont_allow_any_single_char=False)
         _test_char_list(self.group_add, 'Job.request.docker.group_add', dont_allow_any_single_char=False)
 
+    @property
+    def tmpfs_memory_gb(self) -> int:
+        res = 0
+        for s in self.storage:
+            if s.is_tmpfs and s.size_gb:
+                res += s.size_gb
+        return res
+
+    @property
+    def tmpfs_memory_mb(self) -> int:
+        # while GB is sometimes 1000^3 instead of 1024^3 (typically HD size etc), for memory, base 1024 is always used.
+        # (and docker seems to use 1000 based, which is silly)
+        return self.tmpfs_memory_gb * 1024
+
+    @property
+    def tmpfs_memory_byte(self) -> int:
+        # while GB is sometimes 1000^3 instead of 1024^3 (typically HD size etc), for memory, base 1024 is always used.
+        # (and docker seems to use 1000 based, which is silly)
+        return self.tmpfs_memory_gb * 1024 * 1024 * 1024
+
     @property
     def image_nopass(self) -> str:
         res = self.image
@@ -584,6 +686,22 @@ class JobRequest:
     scheduling: JobRequestScheduling = field(default_factory=JobRequestScheduling)
     extra: JobRequestExtra = field(default_factory=JobRequestExtra)
 
+    @property
+    def all_cpu_memory_gb(self) -> int:
+        return (self.resources.cpu_memory_gb or 0) + (self.docker.tmpfs_memory_gb or 0)
+
+    @property
+    def all_cpu_memory_mb(self) -> int:
+        # while GB is sometimes 1000^3 instead of 1024^3 (typically HD size etc), for memory, base 1024 is always used.
+        # (and docker seems to use 1000 based, which is silly)
+        return self.all_cpu_memory_gb * 1024
+
+    @property
+    def all_cpu_memory_byte(self) -> int:
+        # while GB is sometimes 1000^3 instead of 1024^3 (typically HD size etc), for memory, base 1024 is always used.
+        # (and docker seems to use 1000 based, which is silly)
+        return self.all_cpu_memory_gb * 1024 * 1024 * 1024
+
     def sanitized_copy(self, logged_in=True, same_project=False) -> 'JobRequest':
         """copy with removed confidential data
         :param logged_in: sanitized copy for logged in user, or for anonymous user?
