@@ -214,7 +214,7 @@ def convert_job1_to_job2(job: Job1) -> Job2:
             working_dir=job.nvdocker_data.working_dir,
             storage=[
                 # note: Job1 JobDataLocation protocol and share_host are ignored -> never used anyway
-                Job2Storage(container_path=dl.mount_point, host_path=dl.share_path)
+                Job2Storage(container_path=dl.mount_point, host_path=dl.share_path, size_gb=None)
                 for dl in job.nvdocker_data.job_data_locations
             ] if job.nvdocker_data.job_data_locations else [],
             port_mappings=[
